"""Micromed module to load and transform data from Micromed recordings to mne format"""
from datetime import timezone
from pathlib import Path
from typing import List, Union

import mne
import numpy as np
from micromed_io.header import MICROMED_ACQ_EQUIPMENT
from micromed_io.trc import MicromedTRC


def create_mne_from_micromed_recording(
    recording_file: Union[str, Path],
    sub_channels: List[str] = None,
    ch_types: Union[List[str], str] = "eeg",
) -> mne.io.RawArray:
    """Create a mne Raw instance from a Micromed recording

    Parameters
    ----------
    recording_file : Union[str, Path]
        The micromed recording file
    sub_channels : List[str], optional
        The channels to pick from the recording. If None, all channels are picked.
    ch_types: Union[List[str], str], optional
        The list of channel types. Types must be in ``['grad', 'mag', 'ref_meg', 'eeg',
        'seeg', 'dbs', 'ecog', 'eog', 'emg', 'ecg', 'resp', 'bio', 'misc', 'stim', 'exci',
        'syst', 'ias', 'gof', 'dipole', 'chpi', 'fnirs_cw_amplitude', 'fnirs_fd_ac_amplitude',
        'fnirs_fd_phase', 'fnirs_od', 'hbo', 'hbr', 'csd']``

    Returns
    -------
    mne.io.RawArray
        A mne Raw instance containing the requested channels

    Notes
    -----
    Some info are hardcoded, such as:

    - ``device_info`` type: Micromed
    - ``device_info`` site: Unknown

    Update the code if needed

    Examples
    --------
    >>> from micromed_io.to_mne import create_mne_from_micromed_recording
    >>> mne_raw = create_mne_from_micromed_recording("path/to/file.TRC")

    """
    micromed_trc = MicromedTRC(recording_file)

    if sub_channels is None:
        sub_channels = micromed_trc.micromed_header.ch_names
    sub_eegs = micromed_trc.get_data(picks=sub_channels)

    info = mne.create_info(
        ch_names=sub_channels,
        sfreq=micromed_trc.sfreq,
        ch_types=ch_types,
    )
    info["device_info"] = {
        "type": "Micromed",
        "model": MICROMED_ACQ_EQUIPMENT[micromed_trc.micromed_header.acq_unit]
        if micromed_trc.micromed_header.acq_unit in MICROMED_ACQ_EQUIPMENT
        else str(micromed_trc.micromed_header.acq_unit),
        "site": "Unknown",
    }
    info["subject_info"] = {
        "his_id": micromed_trc.micromed_header.name,
        "sex": 0,
    }

    raw = mne.io.RawArray(
        sub_eegs,
        info,
    )
    raw.set_meas_date(
        micromed_trc.micromed_header.recording_date.replace(tzinfo=timezone.utc)
    )

    # Add annotations from notes
    for note_sample, note_val in micromed_trc.get_notes().items():
        raw.annotations.append(
            onset=note_sample / info["sfreq"], duration=0, description=note_val
        )

    # Add annotations from markers
    for marker_sample, marker_val in micromed_trc.get_markers().items():
        raw.annotations.append(
            onset=marker_sample / info["sfreq"], duration=0, description=marker_val
        )

    return raw



"""
Emulate from a TRC file Micromed sending data through TCP.

Note: The server must be launched before.

HOW TO USE:
> python emulate_trc_tcpip.py --help
> python emulate_trc_tcpip.py --file=../data/sample.TRC
"""
import socket
from datetime import datetime
import time
import logging
from pathlib import Path
import click

from .in_out import MicromedIO

PACKET_TIME = 64  # ms


@click.command(context_settings=dict(max_content_width=120))
@click.option("--file", "-f", type=str, required=True, help="the TRC file to emulate")
@click.option(
    "--address",
    "-a",
    default="localhost",
    type=str,
    required=False,
    help="the TCP address to connect to",
    show_default=True,
)
@click.option(
    "--port",
    "-p",
    default=5123,
    type=int,
    required=False,
    help="The TCP port number to connect to",
    show_default=True,
)
@click.option(
    "--verbosity",
    "-v",
    default="1",
    type=click.Choice(["0", "1", "2"]),
    required=False,
    help="Increase output verbosity",
    show_default=True,
)
def run(file: str, address: str, port: int, verbosity: int) -> None:
    """Emulate a Micromed TCP client based on a TRC file"""
    logging.basicConfig(level=0, format=("%(asctime)s\t\t%(levelname)s\t\t%(message)s"))
    verbosity = int(verbosity)  # because of click choice...
    # check if trc file exists
    if not Path(file).exists():
        raise FileNotFoundError(f"{file} does not exist")

    # open trc file
    with open(str(Path(file)), "rb") as f:
        b_data_trc = f.read()
    if verbosity >= 1:
        logging.info("TRC file read done.")

    # create Micromed
    micromed_io = MicromedIO()
    micromed_io.decode_data_header_packet(b_data_trc)

    # Create a TCP/IP socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # Bind the socket to the port
    server_address = (address, port)
    if verbosity >= 1:
        logging.info("starting up on %s port %s" % server_address)
    is_not_connected = True

    while is_not_connected:
        try:
            sock.connect(server_address)
            is_not_connected = False
        except Exception as e:
            logging.warning(e)

    # some general variables
    sample_length = (
        micromed_io.micromed_header.nb_of_channels
        * micromed_io.micromed_header.nb_of_bytes
    )
    sfreq = micromed_io.micromed_header.min_sampling_rate
    n_samples_per_packet = int(PACKET_TIME * 1e3 / sfreq)
    packet_length = sample_length * n_samples_per_packet
    s = micromed_io.micromed_header.data_address
    if verbosity >= 1:
        logging.info("Connected!")
        logging.info(
            f"Sending {len(b_data_trc[s:]) // sample_length} samples for {len(b_data_trc[s:]) / (sample_length * sfreq)} sec"
        )

    # Send the Micromed header data
    tcp_header = bytearray(b"MICM")
    tcp_header.extend((0).to_bytes(2, byteorder="little"))
    tcp_header.extend(s.to_bytes(4, byteorder="little"))
    if verbosity >= 1:
        logging.info("tcp header for header data")
        logging.info(tcp_header)
    sock.send(tcp_header)
    sock.send(b_data_trc[:s])  # send micromed header
    time.sleep(0.1)  # give time for the header to be sent

    start_time = datetime.now()
    current_data_sample = 0
    while True:
        if (current_data_sample + n_samples_per_packet) / sfreq <= (
            datetime.now() - start_time
        ).total_seconds():
            current_data_sample += n_samples_per_packet
            try:
                tcp_header = bytearray(b"MICM")
                tcp_header.extend((1).to_bytes(2, byteorder="little"))
                tcp_header.extend(packet_length.to_bytes(4, byteorder="little"))
                sock.send(tcp_header)
                time.sleep(1e-3)
                sock.send(
                    b_data_trc[
                        s
                        + current_data_sample * (packet_length) : s
                        + (current_data_sample + 1) * packet_length
                    ]
                )

            except Exception as e:
                logging.error("exception catched: " + str(e))
                break

            # debug only
            if verbosity == 2:
                logging.info(f"sending bytes")

            # check if no more data
            if (
                current_data_sample + n_samples_per_packet
                >= len(b_data_trc[s:]) / sample_length
            ):
                logging.info("TRC file over")
                break

    # Clean up the connection
    logging.info("Closing the server")
    sock.close()


if __name__ == "__main__":
    run()



"""Micromed TCP module."""

from enum import IntEnum
import sys

sys.path.append("..")
from micromed_io.in_out import MicromedIO
import logging


class MicromedPacketType(IntEnum):
    """Micromed packet type."""

    HEADER = 0
    EEG_DATA = 1
    NOTE = 2


def decode_tcp_header_packet(packet: bytearray):
    """Decode the Micromed tcp header packet.

    Parameters
    ----------
    packet : bytearray
        The tcp packet to decode.

    Returns
    -------
    bool
        True if decoding was successful. Else False.
    """
    packet_type = None
    next_packet_size = None
    if len(packet) > 0:
        # Check that the packet is not corrupted
        if packet[:4].decode() == "MICM":
            packet_type = int.from_bytes(packet[4:6], "little")
            next_packet_size = int.from_bytes(packet[6:10], "little")
        else:
            logging.warning(f"Wrong header packet: {packet.decode()}")
    else:
        logging.warning(f"header empty packet: [{packet.decode()}]")
    return packet_type, next_packet_size



"""Micromed IO module
"""

from typing import List
import logging
import numpy as np

from micromed_io.in_out import MicromedIO


class MicromedBuffer(MicromedIO):
    # pylint: disable=line-too-long
    """Micromed IO class

    This class TODO

    Parameters
    ----------
    epoch_duration: float, default=None
        The epoch duration in sec.
    epoch_overlap: float, default=None
        The epoch overlap in sec.

    Attributes
    ----------
    epoch_duration: float
        The epoch duration in sec.
    epoch_overlap: float
        The epoch overlap in sec.
    epoch_buffer: np.ndarray
        The epoch buffer that is filled every new eeg packet received. (see Notes 2.)
    current_buffer_length: int
        The position of the next eeg data to fill. (see Notes 1.)
    current_epoch: np.ndarray
        The last full epoch.

    Notes
    -----
    #. The epoch buffer works this way:

        - each time we receive new data, we fill the buffer with the selected channels (cf ``picks``)
        - we keep the position of the "filled data" in current_buffer_length
        - as soon as an epoch_buffer is full, the array is stored in current_epoch and the new data \
    are filled back from current_buffer_length=0.

    #. Be careful with ``picks`` order as it changes the epoch structure (order from ``picks`` is kept). \
    ``["Fp1-G2", "Fpz-G2", "MKR+-MKR-"]`` doest not give the same epoch buffer as \
    ``["MKR+-MKR-", "Fp1-G2", "Fpz-G2"]``.


    See also
    --------
    micromed_io.header.MicromedHeader

    """

    def __init__(
        self,
        epoch_duration: float = None,
        epoch_overlap: float = None,
        picks: List[str] = None,
    ):
        MicromedIO.__init__(self, picks)
        self.epoch_duration = epoch_duration
        self.epoch_overlap = epoch_overlap
        self.epoch_buffer = None
        self.current_epoch = None
        self.current_buffer_length = 0  # count number of added eegs in buffer

    def init_buffer(self) -> None:
        """Init the epoch buffer

        Raises
        ------
        ValueError
            If `epoch_overlap` is more than 0 as it is not supported yet. Feel free to update the code.
        """
        # Construct the epoch buffer
        nb_of_samples = int(
            self.micromed_header.min_sampling_rate * self.epoch_duration
        )
        logging.debug(
            f"Creating an epoch buffer of shape [{len(self.picks_id)}, {nb_of_samples}]"
            + f". picks_id={self.picks_id}"
        )

        self.epoch_buffer = np.empty((len(self.picks_id), nb_of_samples))
        self.current_epoch = np.empty((len(self.picks_id), nb_of_samples))

    def update_epoch_buffer(self) -> bool:
        """Update of the epoch buffer

        The epoch buffer works this way:

            #. each time we receive new data, we fill the buffer
            #. we keep the position of the "filled data" in ``current_buffer_length``
            #. as soon as an ``epoch_buffer`` is full, the array is stored in ``current_epoch`` and\
            the new data are filled back from ``current_buffer_length=0``.

        Returns
        -------
        bool
            If ``True``, a new epoch has been filled and stored in ``self.current_epoch``.

        Raises
        ------
        ValueError
            If the *buffer* is not initialized before this function is called.
        """
        if self.epoch_buffer is None:
            self.init_buffer()

        has_new_epoch = False
        size_to_add = self.current_data_eeg.shape[1]
        n_overlaping_samples = int(
            self.micromed_header.min_sampling_rate * self.epoch_overlap
        )

        # if we have less new data than epoch buffer need to be filled
        if self.current_buffer_length + size_to_add < self.epoch_buffer.shape[1]:
            self.epoch_buffer[
                :, self.current_buffer_length : self.current_buffer_length + size_to_add
            ] = self.current_data_eeg[self.picks_id, :]
            self.current_buffer_length += size_to_add

        # if we have the exact number of new data needed to fill the epoch buffer
        elif self.current_buffer_length + size_to_add == self.epoch_buffer.shape[1]:
            self.epoch_buffer[
                :, self.current_buffer_length : self.current_buffer_length + size_to_add
            ] = self.current_data_eeg[self.picks_id, :]
            self.current_epoch = np.copy(self.epoch_buffer)
            self.current_buffer_length = n_overlaping_samples
            # copy the last samples to the buffer's begining (overlaping window)
            self.epoch_buffer[:, :n_overlaping_samples] = np.copy(
                self.epoch_buffer[
                    :, (self.epoch_buffer.shape[1] - n_overlaping_samples) :
                ]
            )
            has_new_epoch = True

        # if we got data overlapping the epoch buffer...
        else:
            # compute the remaining size (in buffer) and
            # over size (number of new data samples exceeding buffer)
            remaining_size = self.epoch_buffer.shape[1] - self.current_buffer_length
            over_size = size_to_add - remaining_size

            # fill the epoch buffer to construct one epoch and store it
            self.epoch_buffer[:, self.current_buffer_length :] = self.current_data_eeg[
                self.picks_id, :remaining_size
            ]
            self.current_epoch = np.copy(self.epoch_buffer)
            has_new_epoch = True

            # copy the last samples to the buffer's begining (overlaping window)
            self.epoch_buffer[:, :n_overlaping_samples] = np.copy(
                self.epoch_buffer[
                    :, (self.epoch_buffer.shape[1] - n_overlaping_samples) :
                ]
            )
            # fill the epoch buffer with the remaining samples from the new data
            self.epoch_buffer[
                :, n_overlaping_samples : (n_overlaping_samples + over_size)
            ] = np.copy(self.current_data_eeg[self.picks_id, remaining_size:])
            self.current_buffer_length = over_size

        return has_new_epoch



"""Micromed IO module
"""

from typing import List
from typing import Union
from pathlib import Path
import logging
import numpy as np

from micromed_io.in_out import MicromedIO


class MicromedTRC(MicromedIO):
    # pylint: disable=line-too-long
    """Micromed TRC class

    This class deals with Micromed TRC file

    Parameters
    ----------
    filename: str or Path
        The TRC filename.

    Attributes
    ----------
    filename: str or Path
        The TRC filename.

    """

    def __init__(
        self,
        filename: Union[str, Path],
    ):
        MicromedIO.__init__(self, None)
        self.filename = filename
        with open(self.filename, "rb") as f:
            b_data = f.read(
                self._get_data_address()
            )  # trick to not open the whole file
            self.decode_data_header_packet(b_data)

    def _get_data_address(self):
        """Read and return data address"""
        with open(self.filename, "rb") as f:
            packet = f.read(142)
            data_address = int.from_bytes(packet[138:142], "little")
        return data_address

    def get_header(self):
        """Get the header"""
        return self.micromed_header

    def get_sfreq(self):
        """Get the sampling frequency"""
        return self.sfreq

    def get_notes(self):
        """Get the notes"""
        return self.micromed_header.notes

    def get_markers(self):
        """Get the markers"""
        return self.micromed_header.markers

    def get_data(
        self, picks: List = None, keep_raw: bool = False, use_volt: bool = False
    ):
        """Get channels data in format (n_channels, n_sample)

        Parameters
        ----------
        picks : List, optional
            A list of channel to extract. If None, all channels are extracted.
            The default is None.
        keep_raw : bool, optional
            If True, the data won't be converted to voltage. The default is False.
        use_volt : bool, optional
            If True, the data is scaled to Volts. If False, whatever unit is used by Micromed.
            Note that you may loose resolution by doing that. The default is False.
        """
        with open(self.filename, "rb") as f:
            f.seek(self._get_data_address())
            b_data = f.read()
            self.decode_data_eeg_packet(b_data, picks, keep_raw, use_volt)
        return self.current_data_eeg



"""Micromed Header module
Contains some description of the micromed header of type 4
"""
from dataclasses import dataclass
from datetime import datetime

MICROMED_ACQ_EQUIPMENT = {
    0: "BQ124 - 24 channels headbox, Internal Interface",
    1: "MS40 - Holter recorder",
    2: "BQ132S - 32 channels headbox, Internal Interface",
    6: "BQ124 - 24 channels headbox, BQ CARD Interface",
    7: "SAM32 - 32 channels headbox, BQ CARD Interface",
    8: "SAM25 - 25 channels headbox, BQ CARD Interface",
    9: "BQ132S R - 32 channels reverse headbox, Internal Interface",
    10: "SAM32 R - 32 channels reverse headbox, BQ CARD Interface",
    11: "SAM25 R - 25 channels reverse headbox, BQ CARD Interface",
    12: "SAM32 - 32 channels headbox, Internal Interface",
    13: "SAM25 - 25 channels headbox, Internal Interface",
    14: "SAM32 R - 32 channels reverse headbox, Internal Interface",
    15: "SAM25 R - 25 channels reverse headbox, Internal Interface",
    16: "SD - 32 channels headbox with jackbox, SD CARD Interface – PCI Internal Interface",
    17: "SD128 - 128 channels headbox, SD CARD Interface – PCI Internal Interface",
    18: "SD96 - 96 channels headbox, SD CARD Interface – PCI Internal Interface",
    19: "SD64 - 64 channels headbox, SD CARD Interface – PCI Internal Interface",
    20: "SD128c - 128 channels headbox with jackbox, SD CARD Interface – PCI Internal Interface",
    21: "SD64c - 64 channels headbox with jackbox, SD CARD Interface – PCI Internal Interface",
    22: "BQ132S - 32 channels headbox, PCI Internal Interface",
    23: "BQ132S R - 32 channels reverse headbox, PCI Internal Interface",
}


@dataclass
class ElectrodeReferences:
    """Electrode references for conversion purpose

    See also
    --------
    MicromedHeader
    """

    logic_min: int = None
    logic_max: int = None
    logic_ground: int = None
    phy_min: int = None
    phy_max: int = None
    units: int = None


@dataclass
class MicromedHeader:
    """Micromed Header data

    The header is common between TRC files and TCP communication. It contains the info to decode
    the data correctly and some experiment info.

    This class extract *some* of the available info.

    Parameters
    ----------
    surname : str, optional
        The patient surname.
    name : str, optional
        The patient name.
    nb_of_channels : int, optional
        The number of channels.
    acq_unit : int, optional
        The acquisition unit
    min_sampling_rate: int, optional
        The minimum sampling rate of EEGs channels.
    nb_of_bytes: int, optional
        The number of bytes on which eeg data is encoded per value. Should be in [1, 2, 4].

        .. warning::
            When selecting a 24-bits resolution in SystemPLUS, values are encoded on 4 bytes...

    header_type: int, optional
        The header type. To be checked for valid parsing.
    stored_channels: int, optional
        Number of stored channels in EEGs data.
    ch_names: list, optional
        The channel names.
    elec_refs: List[ElectrodeReferences], optional
        A list of electrode references.
    data_address: int, optional
        The byte address of data packet start (useless in TCP context)
    note_address : int, optional
        The byte address of note packet start (useless in TCP context)

    Attributes
    ----------
    surname : str
        The patient surname.
    name : str
        The patient name.
    nb_of_channels : int
        The number of channels.
    acq_unit : int, optional
        The acquisition unit
    min_sampling_rate: int
        The minimum sampling rate of EEGs channels.
    nb_of_bytes: int
        The number of bytes on which eeg data is encoded per value. Should be in [1, 2, 4].

        .. warning::
            When selecting a 24-bits resolution in SystemPLUS, values are encoded on 4 bytes...

    header_type: int
        The header type. To be checked for valid parsing.
    stored_channels: int
        Number of stored channels in EEGs data.
    ch_names: list
        The channel names.
    elec_refs: List[ElectrodeReferences]
        A list of electrode references.
    data_address: int
        The byte address of data packet start (useless in TCP context).
    note_address : int
        The byte address of note packet start (useless in TCP context)
    recording_date: datetime
        The date of file creation and therefore the recording date.
    notes : dict
        Notes entered in Micromed interface. Keys are samples and values are comments (str).
    markers : dict
        Serial markers received by Micromed. Key is the sample and value is the marker value (str).
    """

    surname: str = None
    name: str = None
    nb_of_channels: int = None
    order: list = None
    acq_unit: int = None
    min_sampling_rate: int = None
    nb_of_bytes: int = None
    header_type: int = None
    stored_channels: int = None
    ch_names: list = None
    # elec_refs is a list of electrode references. Dim 2 is
    # [logic_min, logic_max, logic_ground, phy_min, phy_max, units]
    elec_refs: list = None
    data_address: int = None
    recording_date: datetime = None
    notes: dict = None
    markers: dict = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['micromed_io']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.7,<9.0.0', 'mne>=1.0.0,<2.0.0']

entry_points = \
{'console_scripts': ['mmio_tcp_emulator = micromed_io.emulate_trc_tcpip:run']}

setup_kwargs = {
    'name': 'micromed-io',
    'version': '0.2.9',
    'description': 'A library to handle Micromed data',
    'long_description': '# Micromed library\n\nLibrary to handle Micromed data. It also provides some useful scripts.\n\n## Install\n\n``` bash\npip install micromed-io\n```\n\n## Convert a Micromed (*.trc*) file to MNE (*.fif*) format\n\n``` python\nfrom micromed_io.to_mne import create_mne_from_micromed_recording\nmne_raw = create_mne_from_micromed_recording("path/to/file.TRC")\n```\n\n## Emulate Micromed TCP from *.trc* file\n\n### CLI tool\n\nUse the following command:\n\n``` bash\nmmio_tcp_emulator --file=../data/sample.TRC --address=localhost --port=5123\nmmio_tcp_emulator --help\n```\n\n### From python script\n\nDownload `emulate_trc_tcpip.py` from the [gihub repo](https://github.com/etiennedemontalivet/micromed-io) in *scripts/*\n\n``` bash\npython emulate_trc_tcpip.py --file=../data/sample.TRC --address=localhost --port=5123\n```\n\nMore details:\n``` bash\npython emulate_trc_tcpip.py --help\n```\n\n## Read and parse Micromed TCP live data\n\nDownload `read_tcp_data.py` from the [gihub repo](https://github.com/etiennedemontalivet/micromed-io) in *scripts/*\n``` bash\npython read_tcp_data.py --address=localhost --port=5123\n```\n\n> **Note**: Micromed TCP behaves as a client. If you want to try the emulate/read TCP script, launch the reader first that acts as server, then the emulator. \n\n## Read Micromed TCP in a sliding window buffer\n\nIf you plan to use the Micromed data as input of a decoder, you probably want epochs of format `(n_channels, n_samples)`. Then the ``MicromedBuffer`` class is for you. The script ``read_tcp_to_epoch.py`` show you how to use it (see the ``PROCESS HERE`` comment). It uses a **buffer** that mimics the **sliding window** and triggers each time it is filled.\n\n``` python\nfrom micromed_io.buffer import MicromedBuffer\nmicromed_buffer = MicromedBuffer(epoch_duration=5, epoch_overlap=2.5)\n\n```\n\n## Read TRC file\n\n``` python\nfrom micromed_io.trc import MicromedTRC\nmmtrc = MicromedTRC("sample.TRC")\n```\nThen you have access to the *trc* data:\n``` python\nmmtrc.get_header()\nmmtrc.get_markers()\nmmtrc.get_data()\nmmtrc.get_notes()\n```\n> **Note:** ``get_data()`` might take times because it loads the brain data\n\n## TODO\n\n- [x] Include serial markers parsing\n- [ ] Parse all info from Micromed header\n- [ ] Emulate serial markers + notes\n- [ ] Add tests \n\nPlease feel free to reach me if you want to contribute.\n',
    'author': 'Etienne de MONTALIVET',
    'author_email': 'etienne.demontalivet@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Micromed IO module"""

from datetime import datetime
from typing import List
import logging
import numpy as np
from micromed_io.header import ElectrodeReferences, MicromedHeader
from datetime import datetime
from struct import unpack

from numpy import dtype


class MicromedIO:
    """Micromed IO class

    This class provides the basics for decoding Micromed data. It is common between files and
    TCP communication.

    Parameters
    ----------
    picks : List[str], default=None
        Channels to pick from EEGs data (all channels are stored). Channel names must follow
        the same format as in SystemPLUS Evolution (see EEG Setup Montage). WARNING: see Note 3.
        Example : ["Fp1-G2", "Fpz-G2", "MKR+-MKR-"]
        If None, all channels are included. The default is None.

    Attributes
    ----------
    micromed_header: MicromedHeader
        The corresponding micromed header of this connection.
    current_data_eeg: np.ndarray
        The current eeg data of shape (nb_of_channels, nb_of_samples).
    picks : List[str]
        Channels to pick from EEGs data (all channels are sent via TCP). Channel names follow
        the same format as in SystemPLUS Evolution (see EEG Setup Montage). WARNING: This impact
        the channels order.
        Example : ["Fp1-G2", "Fpz-G2", "MKR+-MKR-"]
    picks_id : np.array
        Indexes of the channels to pick.
    sfreq : float
        The sampling frequency


    See also
    --------
    micromed_io.header.MicromedHeader

    """

    def __init__(
        self,
        picks: List[str] = None,
    ):
        self.micromed_header = MicromedHeader()
        self.current_data_eeg = None
        self.picks = picks
        self.picks_id = None
        self.epoch_buffer = None
        self.mkr_index = None  # index of MKR+-MKR- channel in micromed_header.ch_names
        self.current_channels = None
        self.notes = {}
        self.note_start_offset = -1

    def decode_data_header_packet(self, packet: bytearray) -> None:
        """Decode most of the micromed header data packet.

        Parameters
        ----------
        packet : bytearray
            The header packet to decode.

        Raises
        ------
        ValueError
            If any "fixCode" sent by Micromed is not corresponding (LABCOD, ORDER,...)

        """
        self.micromed_header.surname = packet[64 : (64 + 22)].decode().strip("\x00")
        self.micromed_header.name = packet[86 : (86 + 20)].decode().strip("\x00")
        self.micromed_header.acq_unit = int.from_bytes(packet[134:136], "little")
        self.micromed_header.nb_of_channels = int.from_bytes(packet[142:144], "little")
        self.micromed_header.min_sampling_rate = int.from_bytes(
            packet[146:148], "little"
        )
        self.sfreq = self.micromed_header.min_sampling_rate
        self.micromed_header.recording_date = datetime(
            day=int.from_bytes(packet[128:129], "little"),
            month=int.from_bytes(packet[129:130], "little"),
            year=int.from_bytes(packet[130:131], "little") + 1900,
            hour=int.from_bytes(packet[131:132], "little"),
            minute=int.from_bytes(packet[132:133], "little"),
            second=int.from_bytes(packet[133:134], "little"),
        )
        self.micromed_header.data_address = int.from_bytes(packet[138:142], "little")
        self.micromed_header.nb_of_bytes = int.from_bytes(packet[148:150], "little")
        self.micromed_header.header_type = int.from_bytes(packet[175:176], "little")

        # Check micromed header type (must be 4)
        if self.micromed_header.header_type != 4:
            raise ValueError(
                f"Error: Header is not 4 but {self.micromed_header.header_type}. "
                + "Parsing is inappropriate."
            )

        # get zones position and length
        zones = {}
        for i_zone in range(15):
            zname, pos, length = unpack(
                "8sII", packet[176 + i_zone * 16 : 176 + (i_zone + 1) * 16]
            )
            zname = zname.decode("iso-8859-1").strip()
            zones[zname] = pos, length

        # ORDER
        pos, length = zones["ORDER"]
        self.micromed_header.order = np.frombuffer(
            packet[(pos) : (pos + length)],
            dtype="u2",
            count=self.micromed_header.nb_of_channels,
        )

        # Retrieve stored channels name
        code_start_offset = int.from_bytes(packet[184:188], "little")
        self.micromed_header.stored_channels = []
        for iCh in range(self.micromed_header.nb_of_channels):
            self.micromed_header.stored_channels.append(
                int.from_bytes(
                    packet[
                        code_start_offset + 2 * iCh : code_start_offset + 2 * (iCh + 1)
                    ],
                    "little",
                )
            )
        if packet[192 : 192 + 8].decode("utf-8").strip() != "LABCOD":
            raise ValueError(
                f"[MICROMED IO] Error: {packet[192 : 192 + 8].decode()} must be equal to 'LABCOD'"
            )
        elec_start_offset = int.from_bytes(packet[200:204], "little")
        self.micromed_header.ch_names = []
        for iCh in self.micromed_header.stored_channels:
            pos_elec = (
                packet[
                    elec_start_offset
                    + 128 * iCh
                    + 2 : elec_start_offset
                    + 128 * iCh
                    + 2
                    + 6
                ]
                .decode("utf-8")
                .strip("\x00")
            )
            neg_elec = (
                packet[
                    elec_start_offset
                    + 128 * iCh
                    + 8 : elec_start_offset
                    + 128 * iCh
                    + 8
                    + 6
                ]
                .decode("utf-8")
                .strip("\x00")
            )
            self.micromed_header.ch_names.append(f"{pos_elec}-{neg_elec}")

        # construct the indexes of channels to pick in epoch buffer
        if self.picks is None:
            self.picks_id = np.arange(len(self.micromed_header.stored_channels))
        else:
            # Check that all channels are pickable
            for ch in self.picks:
                if ch not in self.micromed_header.ch_names:
                    raise ValueError(
                        f"[MICROMED IO] {ch} is not in "
                        + f"{self.micromed_header.ch_names}. Please fix it in config.ini file."
                    )
            self.picks_id = np.array(
                [self.micromed_header.ch_names.index(ch) for ch in self.picks],
                dtype=int,
            )

        # Extract electrode references for conversion purpose
        self.micromed_header.elec_refs = []
        for iCh in self.micromed_header.stored_channels:
            refs = ElectrodeReferences()
            start = elec_start_offset + 128 * iCh + 14
            refs.logic_min = int.from_bytes(
                packet[start : start + 4], "little", signed=True
            )
            refs.logic_max = int.from_bytes(
                packet[start + 4 : start + 8], "little", signed=True
            )
            refs.logic_ground = int.from_bytes(
                packet[start + 8 : start + 12], "little", signed=True
            )
            refs.phy_min = int.from_bytes(
                packet[start + 12 : start + 16], "little", signed=True
            )
            refs.phy_max = int.from_bytes(
                packet[start + 16 : start + 20], "little", signed=True
            )
            refs.units = int.from_bytes(
                packet[start + 20 : start + 22], "little", signed=True
            )
            self.micromed_header.elec_refs.append(refs)

        # MARKERS
        pos, length = zones["TRIGGER"]
        dt = dtype([("sample", "u4"), ("code", "u2")])
        markers = np.frombuffer(
            packet[pos : pos + length], dtype=dt, count=int(length / dt.itemsize)
        )
        markers_dict = {}
        for marker_sample, marker_val in markers:
            if marker_sample == 4294967295 and marker_val == 65535:
                break
            markers_dict[marker_sample] = str(marker_val)
        self.micromed_header.markers = markers_dict

        # NOTES
        pos, length = zones["NOTE"]
        dt = dtype([("sample", "u4"), ("text", "S40")])
        notes = np.frombuffer(
            packet[pos : pos + length], dtype=dt, count=int(length / dt.itemsize)
        )
        notes_dict = {}
        for note_sample, note_val in notes:
            if note_sample == 0:
                break
            notes_dict[note_sample] = note_val.decode("utf-8")
        self.micromed_header.notes = notes_dict

    def decode_operator_note_packet(self, note_packet: bytearray):
        """Decode the operator notes
        TODO

        Parameters
        ----------
        note_packet : bytearray
            The packet to decode. It must start at the note area.
        """
        pass

    # pylint: disable=too-many-branches,too-many-statements
    def decode_data_eeg_packet(
        self,
        packet: bytearray,
        picks: List = None,
        keep_raw: bool = False,
        use_volt: bool = False,
    ) -> bool:
        """Decode eeg data packet.
        Conversion is made to get physiological value of eegs.

        Parameters
        ----------
        packet : bytearray
            The EEGs data packet to decode.
        picks : List, optional
            A list of channel to extract. If None, all channels are extracted.
            The default is None.
        keep_raw : bool, optional
            If True, the data won't be converted to voltage. The default is False.
        use_volt : bool, optional
            If True, the data is scaled to Volts. If False, whatever unit is used by Micromed.
            Note that you may loose resolution by doing that. The default is False.

        Returns
        -------
        bool
            False if decoding is suspicious/wrong.

        Raises
        ------
        ValueError
            If data is not encoded on 1,2 or 4 bytes. Should not happen.

        Notes
        -----
        EEGs data are stored in this order (per packet):
        ``Ch1_t0, Ch2_t0, ... ChN_t0, Ch1_t1, Ch2_t1, ... ChN_t1, ...``

        ``self.current_data_eeg`` is always ordered regarding the order of the channels
        in self.micromed_header.ch_names
        """

        packet_size = len(packet)
        n_bytes = self.micromed_header.nb_of_bytes
        nb_of_channels = self.micromed_header.nb_of_channels

        if picks is None:
            picks = self.micromed_header.ch_names

        if n_bytes in [1, 2]:
            dt = np.dtype(np.uint16, "little")
        elif n_bytes == 4:
            dt = np.dtype(np.int32, "little")
        else:
            raise ValueError(f"Error: code not yet implemented for n_bytes={n_bytes}.")

        # MKR+MKR- is used to double check extraction/parsing of data
        # if it is not in picks, we add it temporary
        orig_picks = picks.copy()
        mkr_channels = [ch for ch in self.micromed_header.ch_names if "MKR" in ch]
        for ch in mkr_channels:
            if ch not in picks:
                picks.append(ch)

        raw_values = np.frombuffer(packet, dtype=dt)
        nb_of_samples = int(packet_size // n_bytes // nb_of_channels)

        reshaped_data = []
        t = np.arange(nb_of_samples) * nb_of_channels
        for i in range(nb_of_channels):
            if self.micromed_header.ch_names[i] in picks:
                if keep_raw is True:
                    reshaped_data.append(np.take(raw_values, t + i))
                else:
                    logic_min = self.micromed_header.elec_refs[i].logic_min
                    logic_max = self.micromed_header.elec_refs[i].logic_max
                    logic_ground = self.micromed_header.elec_refs[i].logic_ground
                    phy_min = self.micromed_header.elec_refs[i].phy_min
                    phy_max = self.micromed_header.elec_refs[i].phy_max
                    loc = (phy_max - phy_min) / (logic_max - logic_min + 1)
                    if use_volt is True:
                        unit = self.micromed_header.elec_refs[i].units
                        ratio = 0
                        if unit == -1:
                            ratio = 1e-9
                        elif unit == 0:
                            ratio = 1e-6
                        elif unit == 1:
                            ratio = 1e-3
                        elif unit == 2:
                            ratio = 1
                        else:
                            raise ValueError(
                                f"Cannot convert data to Volts. unit is inappropriate: {unit}."
                            )
                        loc *= ratio
                    reshaped_data.append(
                        np.multiply(
                            np.subtract(
                                np.take(raw_values, t + i).astype(int), logic_ground
                            ),
                            loc,
                        )
                    )

        self.current_data_eeg = np.array(reshaped_data)
        self.current_channels = picks
        if keep_raw is True:
            is_well_extracted = True
        else:
            is_well_extracted = self._check_eegs_data(mkr_channels)

        for ch in mkr_channels:
            if ch not in orig_picks:
                self.current_data_eeg = np.delete(
                    self.current_data_eeg,
                    (self.current_channels.index(ch)),
                    axis=0,
                )
                self.current_channels.remove(ch)

        return is_well_extracted

    def _check_eegs_data(self, mkr_channels) -> bool:
        """Check that eegs data are not corrupted

        Returns
        -------
        bool
            True if MKR+-MKR- channel absolute values are close to 50 (µV). Else False.
        """
        success = True
        for ch in mkr_channels:
            if not all(
                np.isclose(
                    abs(self.current_data_eeg[self.current_channels.index(ch)]), 50
                )
            ):
                success = False
        if not success:
            logging.warning("MKR channel(s) is not close to 50mV")

        return success
